/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: SimpleClientManager.java,v 1.2 2001/03/07 08:07:16 wood Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.PI;

import org.omg.PortableInterceptor.ClientRequestInfo;
import org.omg.PortableInterceptor.ClientRequestInterceptor;
import org.omg.PortableInterceptor.CurrentOperations;

/**
 * This class is the manager for portable interceptor on client side.
 * 
 * @author <a href="mailto:jdaniel@intalio.com">Jerome Daniel &lt;daniel@intalio.com&gt;</a>
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.2 $ $Date: 2001/03/07 08:07:16 $ 
 */
public class SimpleClientManager implements ClientManager 
{			
	/**
	 * The client request interceptor list.
	 */
	private ClientRequestInterceptor [] _list = null;
	
	/**
	 * Current impl. Pushes and pops around the send/recieve functions
	 */
	private CurrentImpl _current;
	
	/**
	 * exception code for exception thrown when an interceptor
	 * gives a non-complient response.
	 */
	public static final int BAD_INTERCEPTOR_RESPONSE = 0;
    //modify 21 jun 2002
    org.openorb.CORBA.SystemExceptionHelper sysExHelp=new org.openorb.CORBA.SystemExceptionHelper();
	
	/**
	 * Constructor
	 */
	public SimpleClientManager(org.omg.PortableInterceptor.ClientRequestInterceptor [] list ,
														 CurrentImpl current)
	{
		_list = list;
		_current = current;
	}
	
	/**
	 * This operation must be called from the client interception point.
	 */
	public void send_request( ClientRequestInfo info, RequestCallback cb )
	{
		int index = 0;
		CurrentOperations table = _current.remove();
    
		// Call 'send_request' on each interceptor
		try
		{
			for (; index<_list.length; index++ ) {
				_list[index].send_request( info );
				_current.remove();
			}
		}
		catch ( org.omg.PortableInterceptor.ForwardRequest ex )
		{
      _current.remove();
      
			// A forward request exception has been received from one of the interceptors.			
			cb.reply_location_forward( ex.forward, ex.permanent );
			
			receive_other_point( index-1, info, cb );
		}
		catch ( org.omg.CORBA.SystemException sex )
		{			
      _current.remove();
      
			// A CORBA System exception has been received from one of the interceptors. 
      /* modified by Diana on Oct. 2001*/
      //if((sex instanceof org.omg.CORBA.REPLY_HERE&&sex.completed!=org.omg.CORBA.CompletionStatus.COMPLETED_YES)||(!(sex instanceof org.omg.CORBA.REPLY_HERE)&&sex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO))
      //			sex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
			
			sex.completed = org.omg.CORBA.CompletionStatus.COMPLETED_NO;
			
			cb.reply_system_exception( sex );			
			
			receive_exception_point( index-1, info, cb );
		}
		catch ( java.lang.RuntimeException ex )
		{
      _current.remove();
      
			// A runtime exception has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::send_request", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );
			
			receive_exception_point( index-1, info, cb );
		}
    catch ( java.lang.Error ex )
    {
      _current.remove();
      
			// An error has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::send_request", "Received an unexcepted exception : " + ex.toString() );

			cb.reply_error( ex );
			
			receive_exception_point( index-1, info, cb );
    }
    _current.set(table);
	}
	
	/**
	 * This operation must be called from the client interception point.
	 */
	public void send_poll(ClientRequestInfo info,RequestCallback cb)
	{
		int index = 0;
		CurrentOperations table = _current.remove();
		
		// Call 'send_poll' on each interceptor
		try
		{
			for (; index<_list.length; index++ ) {
				_list[index].send_poll( info );
				_current.remove();
			}
		}
		catch ( org.omg.CORBA.SystemException sex )
		{
      _current.remove();
      
			// A CORBA System exception has been received from one of the interceptors. 
			
			if(sex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO)
				sex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
			
			cb.reply_system_exception( sex );
			
			receive_exception_point( index-1, info, cb );
		}		
		catch ( java.lang.RuntimeException ex )
		{
      _current.remove();

      // A runtime exception has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::send_poll", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );
			
			receive_exception_point( index-1, info, cb );
		}
    catch ( java.lang.Error ex )
    {
      _current.remove();

      // An error has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::send_poll", "Received an unexcepted exception : " + ex.toString() );

			cb.reply_error( ex );
			
			receive_exception_point( index-1, info, cb );
    }
    _current.set(table);
	}
	
	/**
	 * This operation must be called from the client interception point when a reply is received.
	 */
	public void receive_reply(ClientRequestInfo info,RequestCallback cb)
	{
		int index = _list.length-1;
		CurrentOperations table = _current.remove();
		
		// Call 'receive_reply' on each interceptor
		try
		{
			for (; index>=0; index-- ) {
				_list[index].receive_reply( info );
				_current.remove();
			}
		}
		catch ( org.omg.CORBA.SystemException sex )
		{
      _current.remove();

      // A CORBA System exception has been received from one of the interceptors. 					
      
			if( sex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO)
				sex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_YES);

			cb.reply_system_exception( sex );			

			receive_exception_point( index-1, info, cb );
		}
		catch ( java.lang.RuntimeException ex )
		{
      _current.remove();

      // A runtime exception has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::receive_reply", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );
			
			receive_exception_point( index-1, info, cb );
		}
    catch ( java.lang.Error ex )
    {
      _current.remove();

      // An error has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::receive_reply", "Received an unexcepted exception : " + ex.toString() );

			cb.reply_error( ex );
			
			receive_exception_point( index-1, info, cb );
    }
    _current.set(table);
	}
	
	/**
	 * This operation must be called from the client interception point.
	 */
	public void receive_exception(ClientRequestInfo info,RequestCallback cb)
	{
		CurrentOperations table = _current.remove();
    receive_exception_point( _list.length-1, info, cb );		
    _current.set(table);
	}
	
	/**
	 * This operation is called to apply the 'receive_exception' operation on each client interceptor.
	 */
	private void receive_exception_point( int index, ClientRequestInfo info, RequestCallback cb )
	{
		if(index < 0)
			return;
		
		org.omg.CORBA.CompletionStatus status;
		if(info.reply_status() == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value)
		    //modify 21 jun 2002
			status = sysExHelp.extract(info.received_exception()).completed;
		else
			status = org.omg.CORBA.CompletionStatus.COMPLETED_YES;
			
		// Call 'receive_exception' on each interceptor
		for (; index>=0; index-- )
		{
			try
			{
				_list[index].receive_exception( info );
			}
			catch ( org.omg.CORBA.SystemException sex )
			{
				if(sex.completed != status)
					sex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, status);

				cb.reply_system_exception( sex );          
			}
			catch ( java.lang.RuntimeException ex )
			{
        // A runtime exception has been received from an interceptor

				org.openorb.util.Trace.high("SimpleClientManager::receive_exception", "Received an unexcepted exception : " + ex.toString() );

				cb.reply_runtime_exception( ex );
			}
			catch ( java.lang.Error ex )
			{
        // An error has been received from an interceptor

				org.openorb.util.Trace.high("SimpleClientManager::receive_exception", "Received an unexcepted exception : " + ex.toString() );

				cb.reply_error( ex );
			}
			catch ( org.omg.PortableInterceptor.ForwardRequest ex )
			{
        _current.remove();
        
				if(status != org.omg.CORBA.CompletionStatus.COMPLETED_YES) {
					cb.reply_system_exception(new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_YES));
					continue;
				}
				
				// A FrowardRequest exception has been received from one of the interceptors.						
				cb.reply_location_forward( ex.forward, ex.permanent );

				receive_other_point( index-1, info, cb );
				return;
			}
      _current.remove();      
		}
	}
	
	/**
	 * This operation must be called from the client interception point.
	 */
	public void receive_other(ClientRequestInfo info,RequestCallback cb)
	{
		CurrentOperations table = _current.remove();
    receive_other_point( _list.length-1, info, cb );		
    _current.set(table);
	}
	
	/**
	 * This operation is called to apply the 'receive_exception' operation on each client interceptor.
	 */
	private void receive_other_point( int index, ClientRequestInfo info, RequestCallback cb )
	{
		if(index < 0)
			return;
		
		// Call 'receive_exception' on each interceptor
    try {
      for (; index>=0; index-- )
      {				
        try
        {
          _list[index].receive_other( info );
        }
        catch ( org.omg.PortableInterceptor.ForwardRequest ex )
        {
					// A FrowardRequest exception has been received from one of the interceptors.						
          cb.reply_location_forward( ex.forward, ex.permanent );
        }
        _current.remove();
      }
    }
		catch ( org.omg.CORBA.SystemException sex )
		{
      _current.remove();
        
			// A CORBA System exception has been received from one of the interceptors. 
			
			if(sex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO)
				sex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
						
			cb.reply_system_exception( sex );			
			
			receive_exception_point( index-1, info, cb );
		}
		catch ( java.lang.RuntimeException ex )
		{
      _current.remove();
      
			// A runtime exception has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::send_request", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );
			
			receive_exception_point( index-1, info, cb );
		}
    catch ( java.lang.Error ex )
    {
      _current.remove();
      
			// An error has been received from an interceptor
			
			org.openorb.util.Trace.high("SimpleClientManager::send_request", "Received an unexcepted exception : " + ex.toString() );

			cb.reply_error( ex );
			
			receive_exception_point( index-1, info, cb );
    }
	}
}  
