/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: SimpleServerManager.java,v 1.2 2001/03/07 08:07:16 wood Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.PI;

import org.omg.PortableInterceptor.ServerRequestInfo;
import org.omg.PortableInterceptor.ServerRequestInterceptor;
import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;
import org.omg.PortableInterceptor.CurrentOperations;

/**
 * This interface describes all operations that must be implemented to provide a server request
 * interceptor manager.
 * 
 * @author <a href="mailto:jdaniel@intalio.com">Jerome Daniel &lt;daniel@intalio.com&gt;</a>
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.2 $ $Date: 2001/03/07 08:07:16 $ 
 */
public class SimpleServerManager implements org.openorb.PI.ServerManager
{
	/**
	 * The server request interceptor list.
	 */
	private ServerRequestInterceptor [] _list;
	
	/**
	 * The PI current
	 */
	private CurrentImpl _current;

	/**
	 * minor code of INTERNAL exception thrown when an interceptor gives
	 * and invalid response.
	 */
	public static final int BAD_INTERCEPTOR_RESPONSE = 0;
    //modify 21 jun 2002
    org.openorb.CORBA.SystemExceptionHelper sysExHelp=new org.openorb.CORBA.SystemExceptionHelper();

	/**
	 * Set the interceptors list
	 */
	public SimpleServerManager(ServerRequestInterceptor [] list,	CurrentImpl current)
	{
		_list = list;
		_current = current;
	}
	
	/**
	 * This operation must be called from the server interception point.
	 */	
    public void receive_request_service_contexts( ServerRequestInfo info, RequestCallback cb )
    {
	int index = 0;		
	CurrentOperations table = _current.remove();
	
	try
	    {
		for (; index<_list.length; index++ ) {
		    _list[ index ].receive_request_service_contexts( info );		
		    _current.remove();
		}
	    }
	catch ( org.omg.CORBA.SystemException ex )
	    {
		_current.remove();
		
		// A system exception has been received from an interceptor
		
		// Force completion status
		
		/* Modified on May 24th 2002*/
		
		if((ex instanceof org.omg.CORBA.STOP_REQUESTException && ex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_YES )||(!(ex instanceof org.omg.CORBA.STOP_REQUESTException) && ex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO))
		    ex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		
		
		cb.reply_system_exception( ex );
		
		send_exception_point( index-1, info, cb );
	    }
	catch ( org.omg.PortableInterceptor.ForwardRequest ex )
	    {
		_current.remove();
		
		// A forward exception has been received from an interceptor
		
		cb.reply_location_forward( ex.forward, ex.permanent );
		
		send_other_point( index-1, info, cb );
	    }
	catch ( java.lang.RuntimeException ex )
	    {
		_current.remove();
		
		// A runtime exception has been received from an interceptor
		
		org.openorb.util.Trace.high("ServerInterceptorManager::recieve_request_service_contexts", "Received an unexcepted exception : " + ex.toString() );
		
		cb.reply_runtime_exception( ex );
		
		send_exception_point( index-1, info, cb );
	    }
	catch ( java.lang.Error ex )
	    {
		_current.remove();
		System.out.println("error when reading args "+ex.toString());
		ex.printStackTrace();
		// An error has been received from an interceptor
		
		org.openorb.util.Trace.high("ServerInterceptorManager::recieve_request_service_contexts", "Received an unexcepted exception : " + ex.toString() );
		
		cb.reply_error( ex );
		
		send_exception_point( index-1, info, cb );
	    }
	
	_current.set(table);
    }
    
    /**
     * This operation must be called from the server interception point.
	 */	
	public void receive_request(ServerRequestInfo info,RequestCallback cb)
	{
		int index = 0;
		CurrentOperations table = _current.remove();
		
		try
		{
			for (; index<_list.length; index++ ) {
				_list[ index ].receive_request( info );		
				_current.remove();
			}
		}
		catch ( org.omg.CORBA.SystemException ex )
		{
			_current.remove();

			// A system exception has been received from an interceptor
			
			// Force completion status
			
			if((ex instanceof org.omg.CORBA.STOP_REQUESTException && ex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_YES )||(!(ex instanceof org.omg.CORBA.STOP_REQUESTException) && ex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO))
			    ex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
			
			cb.reply_system_exception( ex );
			
			send_exception_point( index-1, info, cb );
		}
		catch ( org.omg.PortableInterceptor.ForwardRequest ex )
		{
			_current.remove();

			// A forward exception has been received from an interceptor
			
			cb.reply_location_forward( ex.forward, ex.permanent );
			
			send_other_point( index-1, info, cb );
		}
		catch ( java.lang.RuntimeException ex )
		{
			_current.remove();

			// A runtime exception has been received from an interceptor
			
      org.openorb.util.Trace.high("ServerInterceptorManager::recieve_request", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );
			
			send_exception_point( index-1, info, cb );
		}
    catch ( java.lang.Error ex )
    {
			_current.remove();

			// An error has been received from an interceptor
			
      org.openorb.util.Trace.high("ServerInterceptorManager::recieve_request", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_error( ex );
			
			send_exception_point( index-1, info, cb );
    }
    _current.set(table);
	}
	
	/**
	 * This operation must be called from the server interception point.
	 */	
	public void send_reply(ServerRequestInfo info,RequestCallback cb)
	{
		int index = _list.length-1;
		CurrentOperations table = _current.remove();
		
		try
		{
			for (; index>=0; index-- ) {
				_list[ index ].send_reply( info );
				_current.remove();
			}
		}
		catch ( org.omg.CORBA.SystemException ex )
		{
			_current.remove();

			// A system exception has been received from an interceptor
			
			if(ex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_YES)
				ex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_YES);
			
			cb.reply_system_exception( ex );
			
			send_exception_point( index-1, info, cb );
		}		
		catch ( java.lang.RuntimeException ex )
		{
			_current.remove();

			// A runtime exception has been received from an interceptor
			
      org.openorb.util.Trace.high("ServerInterceptorManager::send_reply", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );
			
			send_exception_point( index-1, info, cb );
		}
    catch ( java.lang.Error ex )
    {
			_current.remove();

			// An error has been received from an interceptor
			
      org.openorb.util.Trace.high("ServerInterceptorManager::send_reply", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_error( ex );
			
			send_exception_point( index-1, info, cb );
    }
    _current.set(table);
	}
	
	/**
	 * This operation must be called from the server interception point.
	 */	
	public void send_exception(ServerRequestInfo info,RequestCallback cb)
	{
		CurrentOperations table = _current.remove();
    send_exception_point( _list.length-1, info, cb );
    _current.set(table);
	}
	
	/**
	 * This operation is invoked to activate the 'send_exception' operation on each server interceptor.
	 */
	private void send_exception_point( int index, ServerRequestInfo info, RequestCallback cb )
	{
		if(index<0)
			return;
		
		org.omg.CORBA.CompletionStatus status;
		if(info.reply_status() == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value)
		    //modify 21 jun 2002
			status = sysExHelp.extract(info.sending_exception()).completed;
		else
			status = org.omg.CORBA.CompletionStatus.COMPLETED_YES;

		for (; index>=0; --index) {
			try
			{
				_list[ index ].send_exception( info );
				_current.remove();
			}
			catch ( org.omg.CORBA.SystemException ex )
			{
				_current.remove();

				// A system exception has been received from an interceptor

				if(ex.completed != status)
					ex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, status);

				cb.reply_system_exception( ex );
			}
			catch ( java.lang.RuntimeException ex )
			{
				_current.remove();

				// A runtime exception has been received from an interceptor
				org.openorb.util.Trace.high("ServerInterceptorManager::send_exception", "Received an unexcepted exception : " + ex.toString() );

				cb.reply_runtime_exception( ex );
			}
			catch ( java.lang.Error ex )
			{
				_current.remove();

				// An error has been received from an interceptor
				org.openorb.util.Trace.high("ServerInterceptorManager::send_exception", "Received an unexcepted exception : " + ex.toString() );

				cb.reply_error( ex );
			}
			catch ( org.omg.PortableInterceptor.ForwardRequest ex )
			{
				_current.remove();

				if(status != org.omg.CORBA.CompletionStatus.COMPLETED_YES) {
					cb.reply_system_exception(new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_YES));
					continue;
				}

				cb.reply_location_forward( ex.forward, ex.permanent );

				send_other_point( index-1, info, cb );
				
				return;
			}
		}
	}
	
	/**
	 * This operation must be called from the server interception point.
	 */	
	public void send_other(ServerRequestInfo info,RequestCallback cb)
	{
		CurrentOperations table = _current.remove();
    send_other_point( _list.length-1, info, cb );
    _current.set(table);
	}
	
	/**
	 * This operation is invoked to activate the 'send_other' operation on each server interceptor.
	 */
	private void send_other_point( int index, ServerRequestInfo info, RequestCallback cb )
	{
		if(index < 0)
			return;
		
		try
		{
			try
			{
				for (; index>=0; --index) {
					_list[ index ].send_other( info );
					_current.remove();
				}
			}
			catch ( org.omg.PortableInterceptor.ForwardRequest ex )
			{
				_current.remove();
				
				// A forward exception has been received from an interceptor
				
				cb.reply_location_forward( ex.forward, ex.permanent );						
			}
			
		}
		catch ( org.omg.CORBA.SystemException ex )
		{
			_current.remove();

			// A system exception has been received from an interceptor
			
			if(ex.completed != org.omg.CORBA.CompletionStatus.COMPLETED_NO)
				ex = new org.omg.CORBA.INTERNAL(BAD_INTERCEPTOR_RESPONSE, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
	
			cb.reply_system_exception( ex );
			
			send_exception_point( index-1, info, cb );
		}
    catch ( java.lang.RuntimeException ex )
    {
			_current.remove();

			// A runtime exception has been received from an interceptor
      org.openorb.util.Trace.high("ServerInterceptorManager::send_other", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_runtime_exception( ex );

      send_exception_point( index-1, info, cb );
    }
    catch ( java.lang.Error ex )
    {
			_current.remove();

			// An error has been received from an interceptor
      org.openorb.util.Trace.high("ServerInterceptorManager::send_other", "Received an unexcepted exception : " + ex.toString() );

      cb.reply_error( ex );

      send_exception_point( index-1, info, cb );
    }		
	}       
}
