/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: IIOPServerChannel.java,v 1.12 2001/08/13 03:47:33 wood Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.iiop;


import java.util.Map;
import java.util.HashMap;

import java.util.Iterator;

import org.openorb.net.*;
import org.openorb.io.BufferSource;
import org.openorb.io.StorageBuffer;
import org.openorb.io.MarshalBuffer;

import org.omg.BiDirPolicy.BOTH;
import org.omg.BiDirPolicy.BIDIRECTIONAL_POLICY_TYPE;
import org.omg.BiDirPolicy.BidirectionalPolicy;
import org.omg.BiDirPolicy.BidirectionalPolicyHelper;


/** 
 * IIOP implementation of {@link org.openorb.net.ServerChannel}
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.12 $ $Date: 2001/08/13 03:47:33 $ 
 */
public class IIOPServerChannel implements ServerChannel {

  /** Creates new IIOPServerChannel */
  IIOPServerChannel(ServerManager serverManager, Transport transport,
				IIOPClientProtocol clientProtocol, org.omg.IOP.Codec codec) 
  {
      //debug
      //System.out.println("Is a new channel is created1");
    _serverManager = serverManager;
    _orb = _serverManager.orb();
    
		try {
			Class [] cargs = new Class[] { org.omg.CORBA.ORB.class, org.omg.GIOP.Version.class, MarshalBuffer.class };
			_os_ctor = ((org.openorb.CORBA.ORB)_orb).getLoader().classConstructor("iiop.CDROutputStreamClass", "org.openorb.iiop.CDROutputStream", cargs);
		}
		catch(Exception ex) {
			throw new org.omg.CORBA.INITIALIZE("Unable to initialize output stream constructor");
		}

		_socketQueue = new SocketQueue(transport);
    _socketQueue.setServerChannel(this);

		_clientProtocol = clientProtocol;
		
    _codec = codec;
    
		_syncState = new Object();
		
		org.openorb.util.Trace.high(this, "created");
		//System.out.println("register channel");
		_serverManager.register_channel(this);
  }
    
  /**
   * Constructor for bidir delegates.
   */
  IIOPServerChannel(ServerManager serverManager, IIOPClientChannel peer, org.omg.IOP.Codec codec) {
      //debug
      //System.out.println("Is a new channel is created 2");
    _serverManager = serverManager;    
    _orb = _serverManager.orb();

		try {
			Class [] cargs = new Class[] { org.omg.CORBA.ORB.class, org.omg.GIOP.Version.class, MarshalBuffer.class };
			_os_ctor = ((org.openorb.CORBA.ORB)_orb).getLoader().classConstructor("iiop.CDROutputStreamClass", "org.openorb.iiop.CDROutputStream", cargs);
		}
		catch(Exception ex) {
			throw new org.omg.CORBA.INITIALIZE("Unable to initialize output stream constructor");
		}

		_peer = peer;
    _socketQueue = _peer._socketQueue;
    _socketQueue.setServerChannel(this);
    _isDelegated = true;
    _codec = codec;
		
		_syncState = _peer._syncState;
		
		_codesets = true;
		_tcsc = _peer._tcsc;
		_tcsw = _peer._tcsw;
  }
    
  private org.omg.CORBA.ORB _orb;
  ServerManager _serverManager;

  SocketQueue _socketQueue;
  
  Object _syncState;
  int _state = STATE_CONNECTED;
  boolean _pendClose = false;
  
	private boolean _codesets = false;
	int _tcsc = 0;
	int _tcsw = 0;
	
  private IIOPClientProtocol _clientProtocol = null;
   
  IIOPClientChannel _peer = null;
  private boolean _isDelegated = false;

  Map _activeRequests = new HashMap();
  int _channelAge;

	private Object _syncRecv = new Object();
  private boolean _inRecv = false;

  private org.omg.GIOP.Version _version = new org.omg.GIOP.Version((byte)1, (byte)2);
  
	private java.lang.reflect.Constructor _os_ctor;
	
  private org.omg.IOP.Codec _codec;

  private org.omg.IIOP.ListenPoint [] _pendingBidirEndpoints;
  
  private static final int CONN_SC_TOTAL = 2;
  
  // the service contexts
  private static final int CONN_SC_CODESETS = 0;
  private static final int CONN_SC_BIDIR = 1;
  
  private boolean [] _recvConnSC = new boolean[CONN_SC_TOTAL];
  
	public int getTCSC() {
		return _tcsc;
	}
	
	public int getTCSW() {
		return _tcsw;
	}
		
	public String toString() {
		return "ServerChannel: " + _socketQueue.toString();
	}

	public org.omg.CORBA.ORB orb() {
    return _orb;
  }
  
  public int state() {
    if(_isDelegated) {
		  synchronized(_syncState) {
				if(_peer != null)
          return STATE_CONNECTED;
        return STATE_CLOSED;
			}
		}
    return _state;
  }
  
  /**
   * Active request count. This is the number of requests which have not
   * yet been sent or are still expecting a reply. This will return -1 if the
   * channel has been perminently closed.
   */
  public int active_requests() {
		synchronized(_syncState) {
			return _activeRequests.size() + ((_peer != null)?_peer._activeRequests.size():0);
		}
  }
  
  /**
   * Indication of channel age. On the client side this is the request ID of
   * the oldest active request, Since request IDs rise sequentialy this gives
   * an indication of the activity on the channel. On server channels this 
   * will result in a call to peek_request_id if active_requests would be
   * non-zero, otherwise it returns whatever peek_request_id returned the
   * last time active_requests dropped to zero.
   */
  public int channel_age() {
		synchronized(_syncState) {
			if(!_activeRequests.isEmpty() || (_peer != null && !_peer._activeRequests.isEmpty()))
				return RequestIDAllocator.peek_request_id();
			if(_peer != null && _channelAge < _peer._channelAge)
				return _peer._channelAge;
			return _channelAge;
		}
  }

  /**
   * Close the connection. This closes the channel immediatly, all replys to
   * current requests will be discarded. This is a disorderly shutdown.
   */
  public void close() {
    if(_isDelegated) {
			_peer.close(true, new org.omg.CORBA.BAD_INV_ORDER(0, org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
      return;
    }
    
		ServerRequest [] serverRequests;
		ClientRequest [] clientRequests;
    synchronized(_syncState) {
      if(_state == STATE_CLOSED)
        return;
      
      _state = STATE_CLOSED;
      _pendClose = false;
 			
			serverRequests = new ServerRequest[_activeRequests.values().size()];
			_activeRequests.values().toArray(serverRequests);
      _activeRequests.clear();
			
			if(_peer != null) {
				clientRequests = new ClientRequest[_peer._activeRequests.values().size()];
				_peer._activeRequests.values().toArray(clientRequests);
        _peer._activeRequests.clear();
        
        _peer._state = ClientChannel.STATE_CLOSED;
        _peer._closeException = new org.omg.CORBA.BAD_INV_ORDER(0, org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE);
	//debug
	//System.out.println("Is it here that the client channel is set to null 2");
        _peer._peer = null;
        _peer = null;
			} else {
				clientRequests = null;
			}
    }
		
    _serverManager.unregister_channel(this);
      
    _socketQueue.close();
		
		for(int i = 0; i < serverRequests.length; ++i)
			serverRequests[i].client_cancel();

		if(clientRequests != null)
			for(int i = 0; i < clientRequests.length; ++i)
				clientRequests[i].cancel(new org.omg.CORBA.BAD_INV_ORDER("ORB shutdown", 
					org.omg.CORBA.OMGVMCID.value | 4, org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
					
		org.openorb.util.Trace.high(this, "closed");		
  }    
      
  /**
   * Close the connection after all currently processing requests are complete.
   * If there are no currently active requests the channel will close immediatly
   * otherwise if ifActive is true then the channel will close once all 
   * currently active requests have completed, and in the mean time new 
   * incomming requests will be silently discarded. This call returns
   * immediatly, use the state function to interrogate the state while a close
   * is pending.
   */
  public void soft_close(boolean ifActive) {
    if(_isDelegated) {
      if(!ifActive)
        _peer.pause();
      else
        _peer.close(false, new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 4, org.omg.CORBA.CompletionStatus.COMPLETED_NO));
      return;
    }
      
    synchronized(_syncState) {
      if(_state == STATE_CLOSED)
        return;
        
			if(active_requests() > 0) {
        if(ifActive) {
          _pendClose = true;
          if(_peer != null)
            _peer._pendClose = true;
        }
				return;
			}
      
			_state = STATE_CLOSED;
			_pendClose = false;
      
			if(_peer != null) {
        _peer._state = ClientChannel.STATE_PAUSED;
	//debug
	//System.out.println("Is it here that the client channel is set to null 3");
        _peer._peer = null;
        _peer = null;
			}
    }
    
    _serverManager.unregister_channel(this);
    
    // close the connection
    _socketQueue.close();
    
		org.openorb.util.Trace.high(this, "closed");
  }
  
  /**
   * Wait the specified amount of time for an incoming message.
   * @returns false if the channel is closed.
   */
  public boolean recv(int timeout) {
    if(_isDelegated)
      return false;

    synchronized(_syncRecv) {
      if(_inRecv)
        return true;
      _inRecv = true;
    }
    try {
      if(_socketQueue.receive(timeout))
        return true;
      
      close();
      return false;
    }
    finally {
      synchronized(_syncRecv) {
        _inRecv = false;
      }
    }    
  }
  
  /**
   * Donate a thread for recieving messages. This function returns when interrupt
   * is called on the thread or the channel is closed.
   */
  public void run_recv() {
      //System.out.println("run_recv");
      if(_isDelegated){
	  //  System.out.println("is delegated");
	return;
      }
   
    synchronized(_syncRecv) {
	if(_inRecv){
	    //System.out.println("In recev");
	    return;
	}
	_inRecv = true;
    }
    try {
	Thread curr = Thread.currentThread();
	while(!curr.interrupted()){
	    if(!_socketQueue.receive(1000)) {
		close();
		break;
	    }
	}
    }
    finally {
	synchronized(_syncRecv) {
	    _inRecv = false;
	}
    }
  }
    
  /**
	 * Called by SocketQueue. Read off the request header, construct a
	 * server request object, add it to the active request list and return.
	 */
  int process_request(byte minor, CDRInputStream is, byte msg_type, BufferSource source, Object tpAssc) {
      IIOPServerRequest request = null;
      int request_id = 0,i;
      byte [] object_id = null;
      switch(msg_type) {
      case org.omg.GIOP.MsgType_1_1._Request: 
	  String operation;
	  byte sync_scope;
	  org.omg.IOP.ServiceContext[] request_service_contexts;    
	  switch(minor) {
	  case 0:
	  case 1:
	      request_service_contexts = org.omg.IOP.ServiceContextListHelper.read(is);
	      
	      request_id = is.read_ulong();
	      /* org.omg.Messaging.SYNC_WITH_TARGET.value  or   org.omg.Messaging.SYNC_WITH_SERVER.value*/
	      /*Modified from 2 to 0 by Diana on Feb. 6 2002*/
	      sync_scope = (is.read_boolean())?(byte)3:(byte)2;
	      //System.out.println("minor 0 1 synchScope is "+sync_scope);
	      // don't need to skip the reserved bytes,
	      object_id = org.omg.CORBA.OctetSeqHelper.read(is);
	      //System.out.println("Print oid in channel 0 1");
	      /*for(i=0;i<object_id.length;i++)
		System.out.print(object_id[i]+" ");
		System.out.println("****");
	      */
	      operation = is.read_string();
	      //System.out.println("Oper is 01 "+operation);
	      // read and discard any principal
	      org.omg.CORBA.OctetSeqHelper.read(is);
	      request = new IIOPServerRequest(_serverManager, this, tpAssc, request_id, is, object_id, 
					      operation, sync_scope, request_service_contexts, 
					      new org.omg.GIOP.Version((byte)1, minor));
	      
	      checkConnectionSCs(request_service_contexts);
	      
	      if(_recvConnSC == null || _recvConnSC[CONN_SC_CODESETS])
		  is.setCodesets(_tcsc, _tcsw);
	      
	      if(!addRequest(request))
		  return 0;
	      break;
	  case 2:
	      //System.out.println("Minor is 2");
	      request_id = is.read_ulong();
	      
	      request = new IIOPServerRequest(_serverManager, this, tpAssc, request_id, is, 
					      new org.omg.GIOP.Version((byte)1, minor));
	      
	      request.request_source = source;
	      if(!addRequest(request))
		  return 0;
	      
	      sync_scope = is.read_octet();

	      //System.out.println("minor 2 synchScope is "+sync_scope);
	      // skip the reserved
	      is.skip(3);
	      short disp=is.read_short();
	      /*
		if(disp==org.omg.GIOP.KeyAddr.value)
		System.out.println("Key disp");
		else
		System.out.println("Hmm! it is "+disp);
	      */
	      object_id = read_target_address(disp, is);
	      /*System.out.println("Print oid in channel");
		for(i=0;i<object_id.length;i++)
		System.out.print(object_id[i]+" ");
		System.out.println("****");
	      */
	      operation = is.read_string();
	      //System.out.println("Oper is "+operation);
	      request_service_contexts = org.omg.IOP.ServiceContextListHelper.read(is);
	      try {
		  is.alignment(8);
	      }
	      catch(org.omg.CORBA.SystemException ex) {
		  // ignore
	      }
	      
	      // NOTE: if we get more connection SCs which will be relevant here
	      // then we'll have to check the connection SCs.
	      if(!_isDelegated)
		  checkConnectionSCs(request_service_contexts);
	      
	      if(_recvConnSC == null || _recvConnSC[CONN_SC_CODESETS])
		  is.setCodesets(_tcsc, _tcsw);
	      
	      request.init(object_id, operation, sync_scope, request_service_contexts);
	  }
	  source.removeWaitingForBufferListener(null);
	  org.openorb.util.Trace.medium(this, "incoming request #" + request_id);
	  //System.out.println("Enqueue request "+request_id);
	  _serverManager.enqueue_request(request);
	  //System.out.println("REquest "+request_id+" enqueued");
	  break;
      case org.omg.GIOP.MsgType_1_1._LocateRequest:
	  switch(minor) {
	  case 0:
	  case 1:
	      request_id = is.read_ulong();
	      object_id = org.omg.CORBA.OctetSeqHelper.read(is);
	      request = new IIOPServerRequest(_serverManager, this, tpAssc, request_id, 
					      object_id, new org.omg.GIOP.Version((byte)1, minor)); 
	      if(!addRequest(request))
		  return 0;
	      break;
	  case 2:
	      request_id = is.read_ulong();
	      request = new IIOPServerRequest(_serverManager, this, tpAssc, request_id, 
					      new org.omg.GIOP.Version((byte)1, minor)); 
	      
	      request.request_source = source;
	      if(!addRequest(request))
		  return 0;
	      
	      object_id = read_target_address(is.read_short(), is);
	      request.init(object_id);
	      break;
	  }
	  source.removeWaitingForBufferListener(null);
	  org.openorb.util.Trace.medium(this, "incoming locate request #" + request_id);
	  _serverManager.enqueue_request(request);
	  break;
      case org.omg.GIOP.MsgType_1_1._CancelRequest:
	  request_id = is.read_ulong();
	  synchronized(_syncState) {
	      request = (IIOPServerRequest)_activeRequests.remove(new Integer(request_id));
	  }
	  if(request != null) {
	      org.openorb.util.Trace.medium(this, "request #" + request_id + " canceled by client");
	      request.client_cancel();
	  }
	  break;
      }
      //System.out.println("Before return Request id is "+request_id);
      return request_id;
  }
    
    void process_fragment(int req_id, StorageBuffer buf, boolean fragFollows, Object tpAssc) {
	// we have a reply fragment
	IIOPServerRequest request;
	synchronized(_syncState) {
	    request = (IIOPServerRequest)_activeRequests.get(new Integer(req_id));
	}
	if(request != null && request.checkRequestTransportAssoc(tpAssc)) {
	    if(fragFollows)
		org.openorb.util.Trace.medium(this, "request #" + req_id + " request fragment");
	    else
		org.openorb.util.Trace.medium(this, "request #" + req_id + " last request fragment");
	    request.request_source.addLast(buf, !fragFollows);
	}
    }
    
    protected void checkConnectionSCs(org.omg.IOP.ServiceContext[] request_service_contexts) {
	if(_recvConnSC == null)
	    return;
	
	boolean recv = false;
	
	boolean readCSFirst = false;
	int i = 0;
	for(int j = 0; j < request_service_contexts.length; ++j)
	    if(request_service_contexts[j].context_id == org.omg.IOP.TAG_CODE_SETS.value) {
		readCSFirst = true;
		i = j;
		break;
	    }
        
	for(; i < request_service_contexts.length; ++i) {
	    switch(request_service_contexts[i].context_id) {
	    case org.omg.IOP.TAG_CODE_SETS.value:
		if(_recvConnSC[CONN_SC_CODESETS])
		    break;
		// don't bother using a codec
		byte [] cd = request_service_contexts[i].context_data;
		if(cd[0] == 0) {
          // big endian
          _tcsc = ((cd[4]&0xFF) << 24) | ((cd[5]&0xFF) << 16) |
                  ((cd[6]&0xFF) << 8) | (cd[7]&0xFF);
          _tcsw = ((cd[8]&0xFF) << 24) | ((cd[9]&0xFF) << 16) |
                  ((cd[10]&0xFF) << 8) | (cd[11]&0xFF);
        } else {
          // little endian
          _tcsc = ((cd[7]&0xFF) << 24) | ((cd[6]&0xFF) << 16) |
                  ((cd[5]&0xFF) << 8) | (cd[4]&0xFF);
          _tcsw = ((cd[11]&0xFF) << 24) | ((cd[10]&0xFF) << 16) |
                  ((cd[9]&0xFF) << 8) | (cd[8]&0xFF);
        }
        if(_peer != null) {
          _peer._tcsc = _tcsc;
          _peer._tcsw = _tcsw;
        }
        _recvConnSC[CONN_SC_CODESETS] = true;
        recv = true;
        break;
      case org.omg.IOP.BI_DIR_IIOP.value:
        if(_recvConnSC[CONN_SC_BIDIR])
          break;
        try {
          if(_clientProtocol != null) {
            org.omg.CORBA.Any any = _codec.decode_value(request_service_contexts[i].context_data, 
              org.omg.IIOP.BiDirIIOPServiceContextHelper.type());
            _pendingBidirEndpoints = org.omg.IIOP.BiDirIIOPServiceContextHelper.extract(any).listen_points;
          }

          _recvConnSC[CONN_SC_BIDIR] = true;
          recv = true;
        }
        catch(org.omg.CORBA.UserException ex) {
          org.openorb.util.Trace.assertTrue(false, "unexpected exception:" + ex.toString());
        }
        break;
      }
      if(readCSFirst) {
        i = -1;
        readCSFirst = false;
      }
    }
    
    if(recv) {
      for(int j = 0; j < _recvConnSC.length; ++j)
        recv = recv && _recvConnSC[j];
      if(recv)
        _recvConnSC = null;
    }
	}
  
  /**
   * Completes the bidir activation. This is called by the server interceptor.
   */
  void checkBiDirActivation(IIOPServerRequest request) {
    if(_pendingBidirEndpoints != null && _peer == null) {
      BidirectionalPolicy pol = null;
      
      try {
        pol = BidirectionalPolicyHelper.narrow(request.get_server_policy(BIDIRECTIONAL_POLICY_TYPE.value));
      }
      catch(org.omg.CORBA.INV_POLICY ex) {
      }

      if(pol != null && pol.value() == BOTH.value) {
        synchronized(_syncState) {
          if(_state == STATE_CONNECTED && _peer == null) {
            _peer = _clientProtocol.createBidirDelegate(this, _pendingBidirEndpoints);
            _pendingBidirEndpoints = null;
            if(_peer != null)
              org.openorb.util.Trace.high(this, "became bidirectional with peer " + _peer.toString());
          }
        }
      }
    }
  }
	
	/**
	 * Adds request. Returns false if state is closed or will be closed soon
	 */
	private boolean addRequest(IIOPServerRequest request) {
		synchronized(_syncState) {
			// if the channel is closed or is closing we can discard incoming
			// requests silently as they will be cancelled with the close 
			// connection message.
      if(_state == STATE_CLOSED || _pendClose)
				return false;
      			
			_activeRequests.put(new Integer(request.request_id()), request);
		}
		return true;
	}
		
  private byte [] read_target_address(short disposition, CDRInputStream is) {
    byte [] object_id = null;
    switch(disposition) {
    case org.omg.GIOP.KeyAddr.value:
      object_id = org.omg.CORBA.OctetSeqHelper.read(is);
      break;
    case org.omg.GIOP.ProfileAddr.value:
      if(is.read_ulong() != org.omg.IOP.TAG_INTERNET_IOP.value)
        return null;
      is.begin_encapsulation();
      // skip version (octet[2])
      is.skip(2);
      // skip host (string)
      is.skip(is.read_long());
      // skip port (short)
      is.skip(2);
      object_id = org.omg.CORBA.OctetSeqHelper.read(is);
      is.end_encapsulation();
      break;
    case org.omg.GIOP.ReferenceAddr.value:
      int offset = is.read_ulong();
      int len = is.read_ulong();
      for(int i = 0; i < len; ++i) {
        if(i == offset) {
          object_id = read_target_address(org.omg.GIOP.ProfileAddr.value, is);
        } else {
          // skip entire profile
          is.skip(4);
          is.skip(is.read_ulong());
        }
      }
      break;
    }
    return object_id;
  }
  
  // called by server request.
  CDROutputStream create_reply_stream(IIOPServerRequest req, int status) 
  {
      if(!_socketQueue.isOpen()) 
	  throw new org.omg.CORBA.COMM_FAILURE("Client has shutdown", 
					       IIOPMinorCodes.COMM_FAILURE_CLIENT_DIED, org.omg.CORBA.CompletionStatus.COMPLETED_YES);
      
      org.omg.GIOP.Version vers = req.version();
      int req_id = req.request_id();
      org.omg.IOP.ServiceContext [] reply_service_contexts = req.get_reply_service_contexts();
      
      MarshalBuffer buf = new MarshalBuffer(marshallListener, req);
      
      CDROutputStream os;
      try {
	  Object [] args = new Object[3];
	  args[0] = _orb;
	  args[1] = vers;
	  args[2] = buf;
	  os = (CDROutputStream)_os_ctor.newInstance(args);
      }
      catch(Exception ex) {
	  ex.printStackTrace();
	  throw new org.omg.CORBA.INITIALIZE("Unable to create CDROutputStream class");
      }
      
      if(_recvConnSC == null || _recvConnSC[CONN_SC_CODESETS])
	  os.setCodesets(_tcsc, _tcsw);
      new HeaderBlock((byte)org.omg.GIOP.MsgType_1_1._Reply, req_id, os);
      
      switch(vers.minor) {
      case 0:
      case 1:
	  org.omg.IOP.ServiceContextListHelper.write(os, reply_service_contexts);
	  os.write_ulong(req_id);        
	  os.write_ulong(status);
	  break;
      case 2:
	  os.write_ulong(req_id);
	  os.allowFragment();
	  os.write_ulong(status);
	  org.omg.IOP.ServiceContextListHelper.write(os, reply_service_contexts);
	  os.pending_alignment(8);
	  break;
      }
      
      org.openorb.util.Trace.medium(this, "request #" + req_id + " reply stream created. Reply status: " + status);
      
      return os;
  }
    
  // called by server request.
  CDROutputStream create_locate_reply_stream(IIOPServerRequest req, int status) {
    if(!_socketQueue.isOpen()) 
      throw new org.omg.CORBA.COMM_FAILURE("Client has shutdown", 
				IIOPMinorCodes.COMM_FAILURE_CLIENT_DIED, org.omg.CORBA.CompletionStatus.COMPLETED_YES);
    
		org.omg.GIOP.Version vers = req.version();
		int req_id = req.request_id();
		
    MarshalBuffer buf = new MarshalBuffer(marshallListener, req);
    CDROutputStream os;
		try {
			Object [] args = new Object[3];
			args[0] = _orb;
			args[1] = vers;
			args[2] = buf;
			os = (CDROutputStream)_os_ctor.newInstance(args);
		}
		catch(Exception ex) {
			throw new org.omg.CORBA.INITIALIZE("Unable to create CDROutputStream class");
		}
		
    new HeaderBlock((byte)org.omg.GIOP.MsgType_1_1._LocateReply, req_id, os);
    
    os.write_ulong(req_id);
    if(vers.minor == (byte)2)
      os.allowFragment();
    os.write_ulong(status);

		org.openorb.util.Trace.medium(this, "request #" + req_id + " locate reply stream created. Reply status: " + status);

		return os;
  }

  /**
	 * any call to process_request always completes with a call to this
   * function. Should close the channel if a close is pending.
	 */
  void release_request(IIOPServerRequest request) {
      synchronized(_syncState) {
	  int req_id = request.request_id();
	  _activeRequests.remove(new Integer(req_id));
	  org.openorb.util.Trace.medium(this, "request #" + req_id + " completed");
	  
	  if(!_activeRequests.isEmpty())
	      return;
	  
	  _channelAge = RequestIDAllocator.peek_request_id();
      }
      
      // handle pending closes
      if(_pendClose) 
	  soft_close(true);
  }
    
    private MarshalBuffer.Listener marshallListener = new MarshalListener();
    private class MarshalListener implements MarshalBuffer.Listener 
    {
	public void availIncreaced(MarshalBuffer buffer, int available, Object cookie) {
	    if(available > SocketQueue.MAX_FRAG_SIZE) {
		IIOPServerRequest req = (IIOPServerRequest)cookie;
		if(!_socketQueue.send(buffer.fragment(SocketQueue.MAX_FRAG_SIZE), req.get_transport_association())) 
		    throw new org.omg.CORBA.COMM_FAILURE("Client has shutdown", 
							 IIOPMinorCodes.COMM_FAILURE_CLIENT_DIED, org.omg.CORBA.CompletionStatus.COMPLETED_YES);
		org.openorb.util.Trace.medium(IIOPServerChannel.this, "request #" + req.request_id() + " last fragment sent");
	    }
	}
	
	public void bufferClosed(MarshalBuffer buffer, int available, Object cookie) {
	    org.openorb.io.StorageBuffer buf = buffer.lastFragment();
	    IIOPServerRequest req = (IIOPServerRequest)cookie;
	    if(!_socketQueue.send(buf, req.get_transport_association()))
		throw new org.omg.CORBA.COMM_FAILURE("Client has shutdown", 
						     IIOPMinorCodes.COMM_FAILURE_CLIENT_DIED, org.omg.CORBA.CompletionStatus.COMPLETED_YES);
	    org.openorb.util.Trace.medium(IIOPServerChannel.this, "request #" + req.request_id() + " last fragment sent");
	}
	
	public void bufferCanceled(MarshalBuffer buffer, org.omg.CORBA.SystemException ex, Object cookie) {
	    // adjust the completion status to yes, we are sending a reply,
	    // we know the operation has completed. 
	    ex.completed = org.omg.CORBA.CompletionStatus.COMPLETED_YES;
	    
	    throw ex;
	}
    }    
}
