/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: SocketQueue.java,v 1.11 2001/08/13 03:47:33 wood Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.iiop;

import java.net.*;
import java.io.*;

import java.util.LinkedList;
import java.util.Iterator;

import org.openorb.io.StorageBuffer;

import org.openorb.io.BufferSource;
import org.openorb.net.Transport;
import org.openorb.net.TransportAssociationHolder;

import org.omg.CORBA.OctetSeqHolder;
import org.omg.CORBA.IntHolder;

/** 
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.11 $ $Date: 2001/08/13 03:47:33 $ 
 */
class SocketQueue {
  SocketQueue(Transport trans) {
		_trans = trans;
    _open = _trans.isOpen();
  }
      
  public static final int MAX_FRAG_SIZE = 120000;

  /**
   * This should be set to match DEFAULT_SCRAP
   */
  private static final int BUFFER_SIZE = 2048;
  
	private Transport _trans;
  private org.omg.CORBA.ORB _orb;
  
  // sync on this object to change. This is the socket state.
  private IIOPServerChannel _serverChannel = null;
  private IIOPClientChannel _clientChannel = null;
  private boolean _connectionOriginator;
  private boolean _useOddEven = false;

  private Object _syncSend = new Object();
	private boolean _open = false;
	
	private java.lang.reflect.Constructor _osCtor = null;
			
  private LinkedList _incommingQueue = new LinkedList();
  
	private int _req_id_1 = -1;
	private boolean _recieve_reply_1 = true;

	Transport getTransport() {
		return _trans;
	}
	
	public String toString() {
		return _trans.toString();
	}
	
	synchronized void setClientChannel(IIOPClientChannel clientChannel) {
    if(_clientChannel == null) {
      _clientChannel = clientChannel;

      /* if(_clientChannel==null) System.out.println("null 1");
	 else
	 System.out.println("not null 1");
      */
      if(_serverChannel == null) {
        _connectionOriginator = true;
        _orb = _clientChannel.orb();
				
				try {
					final Class [] cargs = new Class[] { org.omg.CORBA.ORB.class, boolean.class, org.omg.GIOP.Version.class, BufferSource.class };
					_osCtor = ((org.openorb.CORBA.ORB)_orb).getLoader().classConstructor("iiop.CDRInputStreamClass", "org.openorb.iiop.CDRInputStream", cargs);
				}
				catch(Exception ex) {
					throw new org.omg.CORBA.INITIALIZE("Unable to create CDROutputStream class");
				}
      } else
        _useOddEven = true;
    }
  }
  
  synchronized void setServerChannel(IIOPServerChannel serverChannel) {
    _serverChannel = serverChannel;
    if(_clientChannel == null) {
      _connectionOriginator = false;
      _orb = _serverChannel.orb();

      try {
        final Class [] cargs = new Class[] { org.omg.CORBA.ORB.class, boolean.class, org.omg.GIOP.Version.class, BufferSource.class };
        _osCtor = ((org.openorb.CORBA.ORB)_orb).getLoader().classConstructor("iiop.CDRInputStreamClass", "org.openorb.iiop.CDRInputStream", cargs);
      }
      catch(Exception ex) {
        throw new org.omg.CORBA.INITIALIZE("Unable to create CDROutputStream class");
      }
    } else
      _useOddEven = true;
  }

  synchronized public boolean isOpen() {
    return _trans.isOpen();
  }
  
  synchronized public void open() {
		_trans.open();
		_open = true;
  }  
  
  synchronized public void close() {
    synchronized(_syncSend) {
      _open = false;
    }
      
		_trans.close();
  }

  // outgoing queue management.  
  
  public boolean send(StorageBuffer buffer, Object assoc) {
		if(!_open)
			return false;
		
    synchronized(_syncSend) {
			if(!_open)
				return false;

      org.openorb.util.Trace.buffer("Sending message", buffer);
      _trans.sendMessage(buffer, assoc);
      
      return true;
    }    
  }
    
  /**
	 * tempory vars. here to avoid creating new every time
	 */
	private OctetSeqHolder tmpBuf = new OctetSeqHolder();
	private IntHolder tmpOff = new IntHolder();	
	private IntHolder tmpLen = new IntHolder();
	
  public boolean receive(int timeout) {
    return process_or_enqueue(true, timeout);
  }
  
  private boolean process_or_enqueue(boolean process, int timeout) {
      int req_id = 0;
      if(process && !_incommingQueue.isEmpty()) {
	  IncommingToProcess rtp = (IncommingToProcess)_incommingQueue.removeFirst();
	  switch(rtp.msg_type) {
	  case org.omg.GIOP.MsgType_1_1._Reply:
	  case org.omg.GIOP.MsgType_1_1._LocateReply:
	      /* if(_clientChannel==null) System.out.println("null 2");
		 else
		 System.out.println("not null 2");
	      */
	      req_id = _clientChannel.process_reply(rtp.minor, rtp.is, rtp.msg_type, rtp.fragFollows, rtp.source, rtp.tpAssc);
	      if(rtp.minor == 1)
		  _req_id_1 = req_id;
	      return true;
	  case org.omg.GIOP.MsgType_1_1._Request:
	  case org.omg.GIOP.MsgType_1_1._LocateRequest:
	  case org.omg.GIOP.MsgType_1_1._CancelRequest:
	      req_id = _serverChannel.process_request(rtp.minor, rtp.is, rtp.msg_type, rtp.source, rtp.tpAssc);
	      if(rtp.minor == 1)
		  _req_id_1 = req_id;
	      return true;
	  case org.omg.GIOP.MsgType_1_1._Fragment:
	      if(rtp.minor == 1)
		  rtp.req_id = _req_id_1;
	      if(rtp.replyFrag)
{
		  /*if(_clientChannel==null) System.out.println("null 3");
		    else
		    System.out.println("not null 3");
		  */
		  _clientChannel.process_fragment(rtp.req_id, rtp.frag, rtp.fragFollows, rtp.tpAssc);
	      }
	      else
		  _serverChannel.process_fragment(rtp.req_id, rtp.frag, rtp.fragFollows, rtp.tpAssc);
	      return true;
	  }
      }
      
      StorageBuffer next_msg = null;
      TransportAssociationHolder trans_assoc = new TransportAssociationHolder();
      
      try {
	  next_msg = _trans.recvMessage(timeout, trans_assoc);
      }
      catch(EOFException ex) {
	  if(_clientChannel != null)
	      _clientChannel.server_pause(new org.omg.CORBA.TRANSIENT(0, org.omg.CORBA.CompletionStatus.COMPLETED_NO));
	  else
{
	      //System.out.println("null 4");
	      _serverChannel.close();
	  }
	  return false;
      }
      catch(org.omg.CORBA.COMM_FAILURE ex) {
	  if(_useOddEven?_connectionOriginator:(_clientChannel != null))
	      _clientChannel.server_pause(ex);
	  else
	      _serverChannel.close();
	  return false;
      }
      
      if(next_msg == null)
	  return true;
      
      org.openorb.util.Trace.buffer("Incoming message", next_msg);
      
      byte minor = -1;
      byte msg_type = -1;
      boolean swap = false;
      boolean fragFollows = false;
      
      // read header information from buffer.
      tmpLen.value = 8;
      int r = next_msg.next(tmpBuf, tmpOff, tmpLen);
      switch(r) {
      case 8:
	  // the normal case..
	  msg_type = tmpBuf.value[tmpOff.value + 7];
	  swap = ((tmpBuf.value[tmpOff.value + 6] & 1) == 1);
	  fragFollows = ((tmpBuf.value[tmpOff.value + 6] & 2) == 2);
	  minor = tmpBuf.value[tmpOff.value + 5];
	  break;
      case 7:
	  swap = ((tmpBuf.value[tmpOff.value + 6] & 1) == 1);
	  fragFollows = ((tmpBuf.value[tmpOff.value + 6] & 2) == 2);
	  // fallthrough
      case 6:
	  minor = tmpBuf.value[tmpOff.value + 5];
	  // fallthrough
      default:
	  // handle shorter reads. This will most likley never occour
	  while(r < 5) {
	      tmpLen.value = 5-r;
	      r += next_msg.next(tmpBuf, tmpOff, tmpLen);
	  }
	  while(r < 8) {
	      tmpLen.value = 1;
	      r += next_msg.next(tmpBuf, tmpOff, tmpLen);
	      switch(r) {
	      case 6:
		  minor = tmpBuf.value[tmpOff.value];
		  break;
	      case 7:
		  swap = ((tmpBuf.value[tmpOff.value] & 1) == 1);
		  fragFollows = ((tmpBuf.value[tmpOff.value] & 2) == 2);
		  break;
	      case 8:
		  msg_type = tmpBuf.value[tmpOff.value];
		  break;
	      }
	  }	
      }
      
      // handle message fragments.
      if(msg_type == org.omg.GIOP.MsgType_1_1._Fragment) {
	  tmpLen.value = 4;
	  org.openorb.util.Trace.assertTrue(next_msg.skip(tmpLen) == 4, "Unable to read header from buffer");
	  
	  boolean replyFrag = true;
	  switch(minor) {
	  case 1:
	      req_id = _req_id_1;
	      replyFrag = _recieve_reply_1;
	      break;
	  case 2: {
				// read request ID. This is copied from the CDRInputStream.
	      int got;		
	      tmpLen.value = 4;
	      if(( got = next_msg.next(tmpBuf, tmpOff, tmpLen)) == 4) {
		  // optimize for normal case.
		  req_id = ((tmpBuf.value[tmpOff.value  ]&0xFF) << ( swap ? 0 : 24 )) |
		      ((tmpBuf.value[tmpOff.value+1]&0xFF) << ( swap ? 8 : 16 )) |
		      ((tmpBuf.value[tmpOff.value+2]&0xFF) << ( swap ? 16 : 8 )) |
		      ((tmpBuf.value[tmpOff.value+3]&0xFF) << ( swap ? 24 : 0 ));
	      } else {
		  // general case. This will probably never occour.
		  req_id = 0;
		  int shf = swap?0:24;
		  while(true) {
		      for(int i = 0; i < got; ++i) {
			  req_id = req_id | ((tmpBuf.value[tmpOff.value+i]&0xFF) << shf);
			  shf += swap?8:-8;
		      }
		      if(swap?(shf < 24):(shf > 0)) {
			  got = next_msg.next(tmpBuf, tmpOff, tmpLen);
			  continue;
		      }
		      break;
		  }
	      }
	      
	      replyFrag = _useOddEven?(((req_id % 2) == 0) == _connectionOriginator):(_clientChannel != null);
	  }	break;
	  }
	  
	  if(process) {
	      if(replyFrag)
{
		  /*if(_clientChannel==null) System.out.println("null 5");
		    else
		    System.out.println("not null 5");
		  */
		  _clientChannel.process_fragment(req_id, next_msg, fragFollows, trans_assoc.value);
}
	      else
		  _serverChannel.process_fragment(req_id, next_msg, fragFollows, trans_assoc.value);
	  } else
	      _incommingQueue.addLast(new IncommingToProcess(req_id, fragFollows, replyFrag, next_msg, trans_assoc.value));
	  return true;
      }
      
      // setup CDRInputStream
      BufferSource source = new BufferSource(next_msg, !fragFollows);
      if(fragFollows) {
	  try {
	      source.addWaitingForBufferListener(waitingForBufferListener);
	  }
	  catch(java.util.TooManyListenersException ex) {
	      org.openorb.util.Trace.assertTrue(false, ex.toString());
	  }
      }
      
      CDRInputStream is;
      try {
	  Object [] args = new Object[4];
	  args[0] = _orb;
	  args[1] = new Boolean(!swap);
	  args[2] = new org.omg.GIOP.Version((byte)1, minor);
	  args[3] = source;
	  is = (CDRInputStream)_osCtor.newInstance(args);
      }
      catch(Exception ex) {
	  throw new org.omg.CORBA.INITIALIZE("Unable to create CDROutputStream class");
      }
      
      // skip over the buffer length (already delt with)
      org.openorb.util.Trace.assertTrue(is.skip(4) == 4, "Unable to read header from buffer");
      
      // look at the message
      
      switch(msg_type) {
      case org.omg.GIOP.MsgType_1_1._Reply:
      case org.omg.GIOP.MsgType_1_1._LocateReply: {
	  if(minor == 1 && fragFollows)
	      _recieve_reply_1 = true;
	  
	  if(process) {
	      /*if(_clientChannel==null) System.out.println("null 6");
		else
		System.out.println("not null 6");
	      */
	      req_id = _clientChannel.process_reply(minor, is, msg_type, fragFollows, source, trans_assoc.value);
	      if(minor == 1 && fragFollows)
		  _req_id_1 = req_id;
	  }	else
	      _incommingQueue.addLast(new IncommingToProcess(msg_type, minor, fragFollows, is, source, trans_assoc.value));
      } break;
      case org.omg.GIOP.MsgType_1_1._Request:
      case org.omg.GIOP.MsgType_1_1._LocateRequest:
      case org.omg.GIOP.MsgType_1_1._CancelRequest: {
	  if(minor == 1 && fragFollows)
	      _recieve_reply_1 = false;
	  
	  if(process) {
	      req_id = _serverChannel.process_request(minor, is, msg_type, source, trans_assoc.value);
	      if(minor == 1 && fragFollows)
		  _req_id_1 = req_id;
	  }	else
	      _incommingQueue.addLast(new IncommingToProcess(msg_type, minor, fragFollows, is, source, trans_assoc.value));
      } break;
      }
      return true;
  }
    
    private BufferSource.WaitingForBufferListener waitingForBufferListener 
	= new BufferSource.WaitingForBufferListener() {
		public boolean waitForBuffer(BufferSource source) {
		    process_or_enqueue(false, 0);
		    return true;
			}
		};	
	  
  private static class IncommingToProcess {
		public IncommingToProcess(byte msg_type) {
			this.msg_type = msg_type;
		}
    public IncommingToProcess(byte msg_type, byte minor, boolean fragFollows, 
															CDRInputStream is, BufferSource source, Object tpAssc) 
    {
      this.minor = minor;
      this.is = is;
      this.msg_type = msg_type;
      this.source = source;
			this.fragFollows = fragFollows;
			this.tpAssc = tpAssc;
    }
		public IncommingToProcess(int req_id, boolean fragFollows, boolean replyFrag, StorageBuffer frag, Object tpAssc) {
      this.msg_type = org.omg.GIOP.MsgType_1_1._Fragment;
			this.req_id = req_id;
			this.fragFollows = fragFollows;
			this.replyFrag = replyFrag;
			this.frag = frag;
			this.tpAssc = tpAssc;
		}
		
    byte minor;
    CDRInputStream is;
    byte msg_type;
    BufferSource source;
		boolean fragFollows;
		Object tpAssc;
		
		int req_id;
		StorageBuffer frag;
		boolean replyFrag;
  }
}
