/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: ClientManager.java,v 1.12 2001/04/12 16:50:46 daniel Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.net;

import java.util.Map;
import java.util.HashMap;

import java.util.ArrayList;
import java.util.Arrays;

import java.util.Iterator;

import org.openorb.policy.*;
import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

/**
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.12 $ $Date: 2001/04/12 16:50:46 $ 
 */
public interface ClientManager {
  /**
   * Return an orb reference.
   */
  public org.omg.CORBA.ORB orb();
  
  /**
   * Get a reference to the server manager. This is used to find the adapter
   * for local requests.
   *
   * @returns the server manager, or null if this is a client-only orb.
   */
  public ServerManager getServerManager();
  
  /**
   * Create bindings for a given IOR. 
   *
   * @param obj target object.
   * @param ior IOR of target.
   * @returns array of client bindings.
   * @throws org.omg.CORBA.INV_OBJREF the object reference is invalid.
   */
    //modify 3 juin 2002
  public ClientBinding [] create_bindings(org.omg.CORBA.Object obj, org.omg.IOP.IOR ior,boolean change);
    
    public ClientBinding [] create_bindings(org.omg.CORBA.Object obj, org.omg.IOP.IOR ior);
	/**
	 * Register a client protocol.
	 */
	public void register_protocol(int profile_tag, ClientProtocol protocol);
  
  /**
   * Register a channel to join the work queue. This is called when the channel
	 * enters the CONNECTED state.
   */
	public boolean register_channel(ClientChannel channel);
	
	/**
	 * Called when a channel exits the CONNECTED state. Returns once
   * all channel threads have completed their work cycles.
	 */
	public void unregister_channel(ClientChannel channel);
  
  /**
   * Shutdown the client side of the orb.
   *
   * @param wait_for_complete true if the operation should not return
   *     until the client side is shut down.
   * @param kill_requests true if requests which are currently being 
   *     processed should be killed.
   */
  public void shutdown(boolean wait_for_complete, boolean kill_requests);
}
